(* Time-stamp: <modified the 18/06/2014 (at 08:46) by Erwan Jahier> *)


(* let put (os: out_channel) (fmt:('a, unit, string, unit) format4) : 'a = *)
(* 	Printf.kprintf (fun t -> output_string os t) fmt *)

open Printf
open Soc2cIdent
open Data

let rec (type_to_string : Data.t -> string -> string) = 
  fun v n -> 
    (* in order to print arrays of arrays type size in the good order, 
       we accumulate the array size when v of of type array, and
       we print it otherwise (finish).
    *)
    let rec finish acc str =
      match acc with
        |[] -> str
        | s::ts -> Printf.sprintf "%s[%d]" (finish ts str) s 
    in 
    let rec aux acc v n =
      match v with
        | Bool -> finish acc ("_boolean "^n)
        | Int -> finish acc ("_integer "^n)
        | Real-> finish acc ("_real "^n)
        | Extern s -> finish acc (s^" "^n)
        | Enum  (s, sl) -> finish acc (id2s s ^" "^n)
        | Struct (sid,_) -> finish acc ((id2s sid)^" "^n)
        | Array (ty, sz) -> aux (sz::acc) ty n
        | Alpha nb -> finish acc ("alpha_"^(string_of_int nb)^" "^n)
        | Alias(a,_) -> finish acc (a^" "^n)
    in
    aux [] v n

let rec (type_to_string2 : Data.t -> string) = 
  fun v -> 
    let str =
      match v with
        | Bool -> "bool"
        | Int -> "int"
        | Real-> "real"
        | Extern s -> s
        | Enum  (s, sl) -> id2s s
        | Struct (sid,_) -> (id2s sid)
        | Array (ty, sz) -> Printf.sprintf "%s_%d" (type_to_string2 ty) sz 
        | Alpha nb -> "alpha_"^(string_of_int nb)
        | Alias(n,_) -> n
    in
    str

(* Soc printer *)
type 'a soc_pp = {
  hfmt:  ('a, unit, string, unit) format4 -> 'a;
  cfmt:  ('a, unit, string, unit) format4 -> 'a;
  cput : string -> unit;
  hput : string -> unit;
  soc: Soc.t
}

let (string_of_soc_key : Soc.key -> string) =
  fun (name,_,_) ->  (id2s name)

let string_of_flow_decl (id, t) = 
  Printf.sprintf "   %s;\n" (type_to_string t (id2s id)) 

let (mem_interface : Soc.t -> string -> bool) =
  fun soc id -> 
    let ins,outs = soc.profile in
    List.mem_assoc id ins || List.mem_assoc id outs

let (is_memory_less : Soc.t -> bool) =
  fun soc -> 
    soc.have_mem = None && soc.instances = []

let rec (string_of_var_expr: Soc.t -> Soc.var_expr -> string) = 
  fun soc -> function
    | Const("true", _) -> "_true"
    | Const("false", _) -> "_false"
    | Const(id, _) -> id2s id
    | Var ("mem_pre",_)   -> (* Clutch! it's not an interface var... *) "ctx->mem_pre" 
    | Var (id,_)   -> 
      if not (mem_interface soc id) then id2s id 
      else if is_memory_less soc then
          sprintf "%s.%s" (get_ctx_name soc.key) (id2s id)
        else 
          sprintf "ctx->%s" (id2s id)
    | Field(f, id,_) -> sprintf "%s.%s" (string_of_var_expr soc f) (id2s id) 
    | Index(f, index,_) -> sprintf "%s[%i]" (string_of_var_expr soc f) index
    | Slice(f,fi,la,st,wi,vt) -> sprintf "%s[%i..%i step %i]; // XXX fixme!\n" 
      (string_of_var_expr soc f) fi la st

open Soc

let (gao2c : Soc.tbl -> 'a soc_pp -> Soc.gao -> unit) =
  fun stbl sp gao -> 
    let string_of_var_expr_list vel =
      let vel = List.map (string_of_var_expr sp.soc) vel in
      String.concat "," vel
    in
    let rec gao2str gao = 
      match gao with
        | Case(id, id_gao_l) -> ( 
          let to_case_str (v,gaol) =
            let gaol_str = (List.map gao2str gaol)@["break;"] in
            let gaol_block = String.concat "      " gaol_str in
            sprintf "\n    case %s:\n   %s" (id2s v) gaol_block
          in
          let cases = List.map to_case_str id_gao_l in
          let str = sprintf "   switch(%s){%s\n   }\n" (id2s id) (String.concat "\n" cases) in
          str
        )
        | Call(vel_out, Assign, vel_in) -> (
          let gen_assign2 vi vo =
            Soc2cUtil.gen_assign (Soc.data_type_of_var_expr vi)
              (string_of_var_expr sp.soc vi) (string_of_var_expr sp.soc vo)
          in
          let l = List.map2 gen_assign2 vel_out vel_in in
          String.concat "" l 
        )
        | Call(vel_out, Method((inst_name,sk),sname), vel_in) -> ( 
          let called_soc = Soc.SocMap.find sk stbl in
          let ctx = Printf.sprintf "ctx->%s" (id2s inst_name) in
          let vel_in = List.map (string_of_var_expr sp.soc) vel_in in
          let vel_out = List.map (string_of_var_expr sp.soc) vel_out in
          Soc2cUtil.gen_step_call sp.soc called_soc vel_out vel_in ctx sname
            ("&ctx->"^(id2s inst_name))
        )
        | Call(vel_out, Procedure sk, vel_in) -> (
          let called_soc = Soc.SocMap.find sk stbl in
          let ctx = get_ctx_name called_soc.key in
          let vel_in = List.map (string_of_var_expr sp.soc) vel_in in
          let vel_out = List.map (string_of_var_expr sp.soc) vel_out in
          Soc2cUtil.gen_step_call sp.soc called_soc vel_out vel_in ctx "step" ""
        )
    in
    sp.cput (gao2str gao)


let (step2c : Soc.tbl -> 'a soc_pp -> Soc.step_method -> unit) =
  fun stbl sp sm -> 
    let sm_str = sm.name in
    let sname = Soc2cUtil.step_name sp.soc.key sm.name in
    let ctx = if is_memory_less sp.soc then "" else
        Printf.sprintf "%s_type* ctx" (get_ctx_name sp.soc.key)
    in
    let ctx_decl = if is_memory_less sp.soc then "" else
        Printf.sprintf "%s_type*" (get_ctx_name sp.soc.key)
    in
    sp.hfmt "void %s(%s);\n" sname ctx_decl;
    sp.cfmt "void %s(%s){\n" sname ctx;
    (match sm.impl with
      | Predef ->        
        sp.cput (SocPredef2c.get sp.soc.key)
      | Gaol(vl, gaol) -> (
        List.iter (fun v -> sp.cput (string_of_flow_decl v)) vl ; 
        sp.cput "\n"; 
        List.iter (gao2c stbl sp) gaol
      )
      | Iterator(it,it_soc_key,s) ->     
        let it_soc = Soc.SocMap.find it_soc_key stbl in
        sp.cput (SocPredef2c.get_iterator sp.soc it it_soc s)
      | Boolred(i,j,k) -> sp.cput (Printf.sprintf "boolred(%i,%i,%i) /* XXX boolred */" i j k)
      | Condact(k,el) -> sp.cput (Printf.sprintf "condact() /* XXX condact */")
    );
    sp.cput (sprintf "\n} // End of %s\n\n" sname)

let (soc2c: int -> out_channel -> out_channel -> Soc.tbl -> Soc.t -> unit) = 
  fun pass hfile cfile stbl soc -> 
	 let hfmt fmt = Printf.kprintf (fun t -> output_string hfile t) fmt in
	 let cfmt fmt = Printf.kprintf (fun t -> output_string cfile t) fmt in
    let hput str = output_string hfile str in
    let cput str = output_string cfile str in
    let put str = cput str; hput str in
    let fmt f str = cfmt f str; hfmt f str in    
    let name, _,_ = soc.key in
    let name = id2s name in
    let il,ol = soc.profile in
    let sp = { hfmt = hfmt; cfmt=cfmt; hput = hput; cput = cput; soc = soc } in
    let ctx_name = get_ctx_name soc.key in
    let ctx_name_type = ctx_name^"_type" in    
    if pass=1 then (
       (* Only for ctx of memoryless nodes + main node *)
      if is_memory_less soc then hfmt "%s %s;\n" ctx_name_type ctx_name;
    ) else (
      if is_memory_less soc then () else (
        cfmt "// Memory allocation for %s\n" ctx_name;
        hfmt "%s_type* %s_new_ctx();\n" ctx_name ctx_name;
        cfmt "%s_type* %s_new_ctx(){" ctx_name ctx_name;
        cfmt "
   %s_type* ctx = (%s_type*)calloc(1, sizeof(%s_type));
   // ctx->client_data = cdata;
   // %s_reset(ctx);
   return ctx;
}
" ctx_name ctx_name ctx_name ctx_name) ;
      

      cfmt "// Step function(s) for %s\n" ctx_name;
      List.iter (step2c stbl sp) soc.step;
      ()
    )

(****************************************************************************)
let (type_to_format_string : Data.t -> string) =
  function
    | Bool -> "%d"
    | Int -> "%d"
    | Real-> "%g"
    | Extern s -> "%s"
    | Enum  (s, sl) -> "%d" 
    | Struct (sid,_) -> "%s"
    | Array (ty, sz) -> "%s"
    | Alpha nb -> assert false

(****************************************************************************)
let rec (lic_type_to_c: Lic.type_  -> string -> string) =
  fun t n -> match t with
    | Bool_type_eff -> "_boolean"^ " " ^ n
    | Int_type_eff  -> "_integer"^ " " ^ n
    | Real_type_eff -> "_real"^ " " ^ n
    | External_type_eff (name) -> (long2s name)^ " " ^ n
    | Abstract_type_eff (name, t) -> lic_type_to_c t n
    | Enum_type_eff (name, l) -> "_integer"^ " " ^ n
    | Array_type_eff (ty, sz) -> 
      Printf.sprintf "%s %s[%d]" (lic_type_to_c ty "") n sz
    | Struct_type_eff (name, fl) ->
      let field_to_c (id,(tf,_opt)) = 
        Printf.sprintf "  %s %s;\n" (lic_type_to_c tf "")  (id2s id)
      in
      (Printf.sprintf "struct %s { %s }" 
        (long2s name) 
        (String.concat "" (List.map field_to_c fl)))^ " " ^ n
    | TypeVar Any -> assert false
    | (TypeVar AnyNum) -> assert false


(****************************************************************************)

let (typedef_of_soc : Soc.t -> string) =
  fun soc -> 
    let ctx_name = get_ctx_name soc.key in
    let ctx_name_type = ctx_name^"_type" in    
    let il,ol = soc.profile in
    let str = Printf.sprintf  "/* %s */\ntypedef struct {\n   /*INPUTS*/\n" ctx_name in
    let str = List.fold_left (fun acc v -> acc^ (string_of_flow_decl v)) str il in
    let str = str ^ "   /*OUTPUTS*/\n" in
    let str = List.fold_left (fun acc v -> acc^ (string_of_flow_decl v)) str ol in
    let str = str ^
      (match soc.have_mem with
        | None -> ""
        | Some t -> 
          Printf.sprintf "   /*Memory cell*/\n   %s ;\n" (id2s (type_to_string t "mem_pre"))
      )
    in
    let str =  str ^ (if soc.instances <> [] then  "   /*INSTANCES*/\n" else "") in
    let string_of_instance (id,sk) = 
      let (sk_id,tl,init_opt) = sk in
      let init = match init_opt with
        | Soc.Nomore -> ""
        | Soc.Slic(_,_,_) -> assert false (* fixme *)
        | Soc.MemInit(ve) -> Printf.sprintf " = %s" (string_of_var_expr soc ve) 
      in
      Printf.sprintf "   %s_type %s%s;\n" (get_ctx_name sk) (id2s id) init
    in
    let str =  List.fold_left (fun acc inst -> acc^(string_of_instance inst)) str soc.instances in
    let str = Printf.sprintf  "%s} %s;\n\n" str ctx_name_type in
    str

module KeySet = Set.Make(struct type t = Soc.key let compare = compare end)

let (typedef : LicPrg.t -> Soc.tbl -> Soc.t -> string) =
  fun licprg soc_tbl main_soc ->
    (* We need to print the ctx typedef a good order
       (w.r.t. typedef dependancies).  To do that, we traverse
       the tree of soc instances which root is the main soc. *)
    let visited = KeySet.empty in
    (* Soc with memory can be used several times; hence we mark via this
       set the ones that have already been visited. *)
    let rec (soc_with_mem : string * KeySet.t -> Soc.t -> string* KeySet.t) =
      fun (acc,visited) soc ->
        if KeySet.mem soc.key visited then (acc,visited) else
          let visited = KeySet.add soc.key visited in
          let acc,visited =
          List.fold_left
            (fun (acc,visited) (iname, sk) ->
              let soc = SocUtils.find_no_exc sk soc_tbl in
              soc_with_mem (acc,visited) soc
            )
            (acc,visited) soc.instances
          in
          let acc = acc ^ (typedef_of_soc soc) in
          acc,visited
    in
    let soc_ctx_typedef_with = 
      if is_memory_less main_soc then "" else fst (soc_with_mem ("",visited) main_soc)
    in
    (* Then we still have to print memoryless soc that can not appear
       as a soc instance *)
    let soc_ctx_typedef_without =
      let socs = Soc.SocMap.bindings soc_tbl in
      let socs = snd (List.split socs) in
      let memless_soc_to_string acc soc =
        if is_memory_less soc then acc^(typedef_of_soc soc) else acc
      in 
      List.fold_left  memless_soc_to_string "" socs
    in 
    (* There are also typedef that comes from user in Lustre V6 *)
    let user_typedef = 
      let to_c k t =
        Printf.sprintf "typedef %s;\n" (lic_type_to_c t (long2s k))
      in
      LicPrg.fold_types (fun k t acc -> acc ^ (to_c k t)) licprg ""
    in
    "// User typedef \n"^user_typedef
    ^"// Memoryless soc ctx typedef \n"^soc_ctx_typedef_without
    ^"// Memoryfull soc ctx typedef \n"^soc_ctx_typedef_with
    
    

(****************************************************************************)
let rec (const_to_c: Lic.const -> string) =
  function
    | Bool_const_eff true -> "1"
    | Bool_const_eff false -> "0"
    | Int_const_eff i -> (sprintf "%s" i)
    | Real_const_eff r -> r
    | Extern_const_eff (s,t) -> (long2s s)
    | Abstract_const_eff (s,t,v,_) -> const_to_c v
    | Enum_const_eff   (s,Enum_type_eff(_,ll)) -> Lic.enum_to_string s ll
    | Struct_const_eff (fl, t) -> (
      let string_of_field = 
        function (id, veff) -> 
          (Ident.to_string id)^" = "^ (const_to_c veff) 
      in
      let flst = List.map string_of_field fl in
(*       (string_of_type_eff t)^ *)
        "{"^(String.concat "; " flst)^"}"
    )
    | Array_const_eff (ctab, t) -> (
      let vl = List.map const_to_c ctab in
      "["^(String.concat ", " vl)^"]"
    )
    | Tuple_const_eff   cl -> assert false


let (constdef : LicPrg.t -> string) = 
  fun licprg -> 
    let to_c k c =
      Printf.sprintf "#define %s %s\n"
(*       Printf.sprintf "const %s = %s;\n" *)
        (long2s k)
        (const_to_c c)
    in
    LicPrg.fold_consts (fun k t acc -> acc ^ (to_c k t)) licprg  "\n// Constant definitions \n"

(****************************************************************************)

let (gen_loop_file : Soc.t -> unit) =
  fun soc -> 
    let base = (string_of_soc_key soc.key) in
    let loopfile = base^"_loop.c" in
    let oc = open_out loopfile in
    let putc s =  output_string oc s in
    let ctx = get_ctx_name soc.key in
    let step = get_step_name soc.key in
    let (n,_,_) = soc.key in
    let n = id2s n in
    Lv6util.entete oc "/*" "*/";
    putc  ("
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include \""^base ^".h\" 
// #include \""^n^".h\"
/* Print a promt ? ************************/
static int ISATTY;
/* MACROS DEFINITIONS ****************/
#ifndef TT
#define TT \"1\"
#endif
#ifndef FF
#define FF \"0\"
#endif
#ifndef BB
#define BB \"bottom\"
#endif
#ifdef CKCHECK
/* set this macro for testing output clocks */
#endif

/* Standard Input procedures **************/
_boolean _get_bool(char* n){
   char b[512];
   _boolean r = 0;
   int s = 1;
   char c;
   do {
      if(ISATTY) {
         if((s != 1)||(r == -1)) printf(\"\a\");
         printf(\"%s (1,t,T/0,f,F) ? \", n);
      }
      if(scanf(\"%s\", b)==EOF) exit(0);
      if (*b == 'q') exit(0);
      s = sscanf(b, \"%c\", &c);
      r = -1;
      if((c == '0') || (c == 'f') || (c == 'F')) r = 0;
      if((c == '1') || (c == 't') || (c == 'T')) r = 1;
   } while((s != 1) || (r == -1));
   return r;
}
_integer _get_int(char* n){
   char b[512];
   _integer r;
   int s = 1;
   do {
      if(ISATTY) {
         if(s != 1) printf(\"\a\");
         printf(\"%s (integer) ? \", n);
      }
      if(scanf(\"%s\", b)==EOF) exit(0);
      if (*b == 'q') exit(0);
      s = sscanf(b, \"%d\", &r);
   } while(s != 1);
   return r;
}
#define REALFORMAT ((sizeof(_real)==8)?\"%lf\":\"%f\")
_real _get_real(char* n){
   char b[512];
   _real r;
   int s = 1;
   do {
      if(ISATTY) {
         if(s != 1) printf(\"\a\");
         printf(\"%s (real) ? \", n);
      }
      if(scanf(\"%s\", b)==EOF) exit(0);
      if (*b == 'q') exit(0);
      s = sscanf(b, REALFORMAT, &r);
   } while(s != 1);
   return r;
}
/* Standard Output procedures **************/
void _put_bottom(char* n){
   if(ISATTY) printf(\"%s = \", n);
   printf(\"%s \", BB);
   if(ISATTY) printf(\"\\n\");
}
void _put_bool(char* n, _boolean _V){
   if(ISATTY) printf(\"%s = \", n);
   printf(\"%s \", (_V)? TT : FF);
   if(ISATTY) printf(\"\\n\");
}
void _put_int(char* n, _integer _V){
   if(ISATTY) printf(\"%s = \", n);
   printf(\"%d \", _V);
   if(ISATTY) printf(\"\\n\");
}
void _put_real(char* n, _real _V){
   if(ISATTY) printf(\"%s = \", n);
   printf(\"%f \", _V);
   if(ISATTY) printf(\"\\n\");
}
/* Output procedures **********************/
#ifdef CKCHECK
void %s_BOT_n(void* cdata){
   _put_bottom(\"n\");
}
#endif
/* Output procedures **********************/
void "^n^"_O_n(void* cdata, _integer _V) {
   _put_int(\"n\", _V);
}/* Main procedure *************************/
int main(){
   
   int s = 0;
   /* Context allocation */
   "^ctx^"_type* ctx = "^(if is_memory_less soc then (
     "&"^ ctx^ ";"
   ) else (ctx^"_new_ctx(NULL);"))^"
   /* Main loop */
   ISATTY = isatty(0);
   first_step=_true;
   while(1){
      if (ISATTY) printf(\"#step %d \\n\", s+1);
      else if(s) printf(\"\\n\");
      fflush(stdout);
      ++s;
");
    let inputs,outputs = soc.profile in
    let inputs = SocVar.expand_profile true true inputs in
    let outputs = SocVar.expand_profile true true  outputs in
    List.iter (fun (id,t) -> 
      let t = type_to_string2 t in
      let str = Printf.sprintf "      ctx->%s = _get_%s(\"%s\");\n" id t id in
      putc str
    ) 
      inputs;
    let inputs_fmt  = List.map (fun (_,t) -> type_to_format_string t) inputs in
    let outputs_fmt = List.map (fun (_,t) -> type_to_format_string t) outputs in
    putc ("      " ^ step^"(ctx);
      printf(\"" ^ (String.concat " " inputs_fmt)^ " #outs " ^ 
             (String.concat " " outputs_fmt)^ "\\n\"," ^
             (String.concat "," (List.map (fun (id,_) -> "ctx->"^id ) (inputs@outputs)))^ 
            ");
      first_step=_false;
   }
   return 1;
   
}
") ;
    flush oc; close_out oc;
    Printf.printf "%s has been generated.\n" loopfile 
    

(****************************************************************************)
(* The entry point for lus2lic --to-c *)
let (f : Lv6MainArgs.t -> Soc.key -> Soc.tbl -> LicPrg.t -> unit) =
  fun args msoc stbl licprg -> 
    let socs = Soc.SocMap.bindings stbl in
    let socs = snd (List.split socs) in 
        (* XXX que fait-on pour les soc predef ? *)
        (*     let _, socs = List.partition is_predef socs in *)
    let base = string_of_soc_key msoc in
    let hfile = base ^  ".h" in
    let cfile = base ^ ".c" in
    let occ = open_out cfile in
    let och = open_out hfile in
    let putc s =  output_string occ s ; flush occ in
    let puth s =  output_string och s ; flush och  in
    let main_soc = Soc.SocMap.find msoc stbl in
    Lv6util.entete occ "/*" "*/" ;
    Lv6util.entete och "/*" "*/";

    gen_loop_file main_soc;
    
    output_string och "
#include <stdlib.h>
#include <string.h>

#ifndef _SOC2C_PREDEF_TYPES
typedef int _boolean;
typedef int _integer;
typedef char* _string;
typedef double _real;
typedef double _double;
typedef float _float;
#define _false 0
#define _true 1
#endif
// _SOC2C_PREDEF_TYPES
";

    putc (Printf.sprintf "#include \"%s\"\n" hfile);
    puth (typedef licprg stbl main_soc );
    putc (constdef licprg);
    puth "/////////////////////////////////////////////////\n";
    puth "//// Static allocation of memoryless soc ctx\n";
    List.iter (soc2c 1 och occ stbl) socs;
    puth "/////////////////////////////////////////////////\n";
    puth "_boolean first_step;\n";
    putc "//// Defining step functions\n";
    List.iter (soc2c 2 och occ stbl) socs;
    flush occ; close_out occ;
    flush och; close_out och;
    Printf.printf "%s has been generated.\n" hfile; 
    Printf.printf "%s has been generated.\n" cfile;
    flush stdout
